import inspect
import io
import logging
import os
import sys
import traceback

DEFAULT_FORMATTER = logging.Formatter(
    "[%(levelname)s] %(name)s %(funcName)s #%(lineno)d -> %(message)s"
)


class PACHandler(logging.Handler):
    def emit(self, record):
        try:
            if record.levelno >= 40:
                stream = sys.stderr
            else:
                stream = sys.stdout

            stream.write(f"{self.format(record)}\n")

            self.acquire()
            try:
                stream.flush()
            finally:
                self.release()
        except RecursionError:
            raise
        except Exception:
            self.handleError(record)


__handler = PACHandler()
__handler.setFormatter(DEFAULT_FORMATTER)


def get_handler():
    return __handler


class PACLogger(logging.Logger):
    def __init__(self, name: str, level: int = logging.INFO):
        super(PACLogger, self).__init__(name=name, level=level)
        self.addHandler(get_handler())

    def findCaller(
        self, stack_info: bool = False, stacklevel: int = 1
    ) -> tuple[str, int, str, str | None]:
        """
        Find the stack frame of the caller so that we can note the source
        file name, line number and function name.
        """
        f = inspect.currentframe()

        rv = "(unknown file)", 0, "(unknown function)", None
        while hasattr(f, "f_code"):
            # work our way up the frame stack to find the first thing that is
            # NOT from core/logging package. This is the real caller (we hope...)
            co = f.f_code
            filename = os.path.normcase(co.co_filename)

            # noinspection PyProtectedMember
            if "core/log" in filename or filename == logging._srcfile:
                f = f.f_back
                continue

            sinfo = None
            if stack_info:
                sio = io.StringIO()
                sio.write("Stack (most recent call last):\n")
                traceback.print_stack(f, file=sio)
                sinfo = sio.getvalue()
                if sinfo[-1] == "\n":
                    sinfo = sinfo[:-1]
                sio.close()

            rv = (co.co_filename, f.f_lineno, co.co_name, sinfo)
            break

        return rv



import argparse
import importlib
import os.path
import sys
from pathlib import Path

from pyside_app_core.theme_generator import compile_qrc_to_resources
from pyside_app_core.style.theme import DEFAULT_THEME


def main():
    parser = argparse.ArgumentParser("compile-pyside-theme")

    parser.add_argument(
        "--extra-python-path",
        help="Provide a directory path to resolve python imports",
    )
    parser.add_argument(
        "--custom-theme",
        help='Provide a python module path for a custom theme object e.g. "my.theme.CustomTheme"',
    )
    parser.add_argument(
        "--extra-qss-templates",
        help='Provide file path to jinja template for qss e.g. "my/theme/templates/custom.qss.jinja2" (Can specify this flag multiple times)',
        action="append",
    )
    parser.add_argument(
        "--extra-qresources",
        help='Provide a python module path for additional QResource object e.g. "my.theme.Resource" (Can specify this flag multiple times)',
        action="append",
    )

    parser.add_argument(
        "target_dir",
        help="Provide a directory path were the final resource file will be placed",
        metavar="target-dir",
    )

    args = parser.parse_args()

    print("...")

    target_dir = Path(args.target_dir)

    if args.extra_python_path:
        pypath = os.path.abspath(args.extra_python_path)
        sys.path.append(pypath)

    theme = DEFAULT_THEME
    if args.custom_theme:
        mod_path, klass_name = args.custom_theme.rsplit(".", 1)

        mod = importlib.import_module(mod_path)

        theme = getattr(mod, klass_name)

    qss_template_extra = None
    if args.extra_qss_templates:
        qss_template_extra = [Path(q) for q in args.extra_qss_templates]

    resources_extra = []
    if args.extra_qresources:
        for rcc in args.extra_qresources:
            mod_path, obj_name = rcc.rsplit(".", 1)

            mod = importlib.import_module(mod_path)

            resources_extra.append(getattr(mod, obj_name))

    generated_file = compile_qrc_to_resources(
        target_dir=target_dir,
        qss_theme=theme,
        qss_template_extra=qss_template_extra,
        resources_extra=resources_extra,
    )

    print(f'Generated: "{generated_file}"')


if __name__ == "__main__":
    main()



from typing import Protocol, Self, TypeVar

T = TypeVar("T")


class CommandInterface(Protocol):
    def encode(self) -> bytearray:
        ...

    @classmethod
    def decode(cls, raw_data: bytes):
        ...


class TranscoderInterface(Protocol):
    @classmethod
    def encode_data(cls, command: CommandInterface) -> bytearray:
        pass

    @classmethod
    def decode_data(cls, data: bytearray) -> CommandInterface:
        pass

    @classmethod
    def format_error(cls, error: Exception) -> CommandInterface:
        pass



from PySide6.QtCore import QPoint, QRect, Qt
from PySide6.QtGui import QAction
from PySide6.QtWidgets import QMainWindow, QWidget

from pyside_app_core.frameless.about_dialog import AboutDialog
from pyside_app_core.frameless.base_window import FramelessBaseMixin
from pyside_app_core.qt.widgets.menu_ctx import MenuBarContext
from pyside_app_core.qt.widgets.tool_bar_ctx import ToolBarContext
from pyside_app_core.qt.widgets.window_settings_mixin import WindowSettingsMixin
from pyside_app_core.services import platform_service


class FramelessMainToolbarWindow(WindowSettingsMixin, FramelessBaseMixin, QMainWindow):
    def __init__(self):
        super(FramelessMainToolbarWindow, self).__init__(parent=None)

        self._about_dialog = AboutDialog()

        self._central = QWidget(parent=self)
        self._central.setStyleSheet("""background-color: none;""")
        self.setCentralWidget(self._central)

        # must call in order to show grab handle
        self.statusBar().show()

        self._menu_bar = MenuBarContext(
            self,
            border_width=(
                0 if platform_service.is_macos else self._theme.win_divider_width
            ),
            border_color=self._theme.win_divider_color,
        )

        self._menu_bar.setNativeMenuBar(platform_service.is_macos)
        self.setMenuWidget(self._menu_bar)

        with self._menu_bar.add_menu(self.tr("File")) as file_menu:
            with file_menu.add_action(self.tr("Quit")) as exit_action:
                exit_action.setMenuRole(QAction.MenuRole.QuitRole)
                exit_action.triggered.connect(self.close)

        with self._menu_bar.add_menu(self.tr("Help")) as help_menu:
            with help_menu.add_action(self.tr("About")) as about_action:
                about_action.setMenuRole(QAction.MenuRole.AboutRole)

                def _show_about():
                    self._about_dialog.exec()

                about_action.triggered.connect(_show_about)

        self._tool_bar = ToolBarContext(area="top", parent=self, movable=False)
        self._tool_bar.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonIconOnly)

        if platform_service.is_macos:
            spacer = QWidget(self)
            spacer.setFixedWidth(self._window_actions.container_width + 5)
            spacer.setFixedHeight(28)
            self._tool_bar.addWidget(spacer)

        self._menu_bar.set_offset(
            5 if platform_service.is_macos else self._window_actions.container_height
        )
        self._menu_bar.set_shift(8)

    @property
    def menu_bar(self):
        return self._menu_bar

    @property
    def tool_bar(self):
        return self._tool_bar

    @property
    def window_bar_geo(self) -> QRect:
        return QRect(QPoint(0, 0), self.tool_bar.geometry().bottomRight())



import logging
from typing import List, Protocol

from PySide6.QtCore import QIODevice, QObject, Qt, QTimer, Signal
from PySide6.QtSerialPort import QSerialPort, QSerialPortInfo

from pyside_app_core import log
from pyside_app_core.errors.serial_errors import (
    SerialConnectionError,
    SerialDisconnectedError,
    SerialError,
    SerialReadError,
    SerialUnknownError,
    SerialWriteError,
)
from pyside_app_core.services.serial_service.utils.abstract_decoder import (
    CommandInterface,
    TranscoderInterface,
)

log.set_level(lvl=logging.DEBUG)


class CanRegister(Protocol):
    def bind_serial_service(self, serial_service: "SerialService"):
        ...


class SerialService(QObject):
    ports = Signal(list)
    com_connect = Signal(QSerialPort)
    com_disconnect = Signal()
    data = Signal(object)
    error = Signal(Exception)

    def __init__(self, transcoder: TranscoderInterface, parent: QObject):
        super(SerialService, self).__init__(parent=parent)

        self._transcoder = transcoder
        self._com: QSerialPort | None = None
        self._buffer = bytearray()

    def _new_com(
        self, port_info: QSerialPortInfo
    ) -> tuple[QSerialPort, QSerialPort.SerialPortError | None]:
        com = QSerialPort(port_info, parent=self)
        com.setBaudRate(QSerialPort.BaudRate.Baud115200)
        open_ok = com.open(QIODevice.OpenModeFlag.ReadWrite)

        return com, None if open_ok else com.error()

    def open_connection(self, port_info: QSerialPortInfo) -> bool:
        self.close_connection()

        self._com, error = self._new_com(port_info)
        if error:
            self._on_error(error)
            return False

        self._com.readyRead.connect(self._on_data)
        self._com.errorOccurred.connect(self._on_error)

        self.com_connect.emit(self._com)

        return True

    def register_reader(self, reader: QObject) -> None:
        if hasattr(reader, "handle_serial_connect"):
            self.com_connect[QSerialPort].connect(
                reader.handle_serial_connect,
                type=Qt.ConnectionType.UniqueConnection,
            )
        if hasattr(reader, "handle_serial_disconnect"):
            self.com_disconnect.connect(
                reader.handle_serial_disconnect,
                type=Qt.ConnectionType.UniqueConnection,
            )
        if hasattr(reader, "handle_serial_ports"):
            self.ports[list].connect(
                reader.handle_serial_ports,
                type=Qt.ConnectionType.UniqueConnection,
            )
        if hasattr(reader, "handle_serial_data"):
            self.data[bytearray].connect(
                reader.handle_serial_data,
                type=Qt.ConnectionType.UniqueConnection,
            )
        if hasattr(reader, "handle_serial_error"):
            self.error[Exception].connect(
                reader.handle_serial_error,
                type=Qt.ConnectionType.UniqueConnection,
            )

    def link(self, *can_register: CanRegister):
        for item in can_register:
            item.bind_serial_service(self)

    def scan_for_ports(self) -> None:
        ports = QSerialPortInfo.availablePorts()

        self._debug_ports(ports)
        log.debug("sending ports...")

        self.ports.emit([p for p in ports if self._port_filter(p)])

    def send_command(self, command: CommandInterface) -> None:
        log.debug(f"sending cmd: {command}")

        if not self._com:
            log.debug("com port not connected")
            return

        self._com.write(command.encode())

    def close_connection(self) -> None:
        if not self._com:
            return

        self.com_disconnect.emit()

        try:
            self._com.errorOccurred.disconnect()
            self._com.readyRead.disconnect()
        except Exception as e:
            log.exception(e)
            pass

        if self._com and self._com.isOpen():
            self._com.flush()
            self._com.close()

        self._com.deleteLater()
        self._com = None

    def deleteLater(self) -> None:
        self.close_connection()
        super().deleteLater()

    def _on_data(self, *args, **kwargs) -> None:
        if not self._com:
            return

        read: bytes = self._com.readAll()
        data = bytearray(read)
        self._buffer.extend(data)

        if b"\x00" in self._buffer:
            chunks: list[bytearray]
            remainder: bytearray
            *chunks, remainder = self._buffer.split(b"\x00")

            for chunk in chunks:
                try:
                    command = self._transcoder.decode_data(chunk)
                except Exception as e:
                    log.exception(e)
                    command = self._transcoder.format_error(e)

                self.data.emit(command)

            self._buffer.clear()
            self._buffer.extend(remainder)

    def _on_error(self, error: QSerialPort.SerialPortError | None) -> None:
        exception: SerialError

        if error is None or error == QSerialPort.SerialPortError.NoError:
            return
        elif error == QSerialPort.SerialPortError.OpenError:
            exception = SerialConnectionError(self._com, error)
        elif error == QSerialPort.SerialPortError.ReadError:
            exception = SerialReadError(self._com, error)
        elif error == QSerialPort.SerialPortError.WriteError:
            exception = SerialWriteError(self._com, error)
        elif error == QSerialPort.SerialPortError.ResourceError:
            exception = SerialDisconnectedError(self._com, error, internal=True)
        else:
            exception = SerialUnknownError(self._com, error)

        self.close_connection()

        QTimer.singleShot(3000, self.scan_for_ports)

        self.error.emit(exception)
        raise exception

    def _port_filter(self, p: QSerialPort) -> bool:
        filters: list[bool] = [
            # p.portName().startswith("cu."),
            # "bluetooth" not in p.portName().lower()
        ]

        return all(filters)

    def _debug_ports(self, ports: List[QSerialPortInfo]) -> None:
        for p in ports:
            log.debug("-----------------------------")
            log.debug(f"name:         {p.portName()}")
            log.debug(f"manufacturer: {p.manufacturer()}")
            log.debug(f"productId:    {p.productIdentifier()}")
            log.debug(f"serialNumber: {p.serialNumber()}")
            log.debug(f"vendorId:     {p.vendorIdentifier()}")
            log.debug(f"systemLoc:    {p.systemLocation()}")



import os
from typing import cast

from PySide6.QtCore import QCoreApplication, QObject, QSettings
from PySide6.QtGui import QShowEvent
from PySide6.QtWidgets import QWidget

from pyside_app_core.qt import application_service


class SettingsMixin:
    def __init__(self, parent: QObject, *args, **kwargs):
        super(SettingsMixin, self).__init__(*args, **kwargs)

        self._settings = QSettings(
            application_service.get_app_id(),
            application_service.get_app_name(),
            parent,
        )

        self._restored = False

        QCoreApplication.instance().aboutToQuit.connect(self._store_state)

    def showEvent(self, event: QShowEvent):
        if not self._restored:
            self._restore_state()
            self._restored = True

        if isinstance(self, QWidget):
            cast(QWidget, super()).showEvent(event)

    def _restore_state(self):
        pass

    def _store_state(self):
        pass



from enum import auto, Enum, IntEnum
from typing import Callable, NamedTuple, Tuple

from PySide6.QtCore import QEvent, QLine, QPoint, QRect, Qt, Signal
from PySide6.QtGui import (
    QBrush,
    QColor,
    QMouseEvent,
    QPainter,
    QPainterPath,
    QPaintEvent,
    QPen,
    QResizeEvent,
)
from PySide6.QtWidgets import QWidget

from pyside_app_core.services import platform_service
from pyside_app_core.qt import application_service


class Action(Enum):
    NONE = auto()
    CLOSE = auto()
    MINIMIZE = auto()
    MAXIMIZE = auto()


ShapeFunc = Callable[[QPainter, QRect, QColor, Action], None]
EnabledFunc = Callable[[], bool]


class ButtonData(NamedTuple):
    color: QColor
    action: Action
    signal: Signal
    shape_fn: ShapeFunc
    enabled_fn: EnabledFunc


class WindowActions(QWidget):
    close_clicked = Signal()
    minimize_clicked = Signal()
    maximize_clicked = Signal()

    def __init__(
        self,
        parent: QWidget,
    ):
        super(WindowActions, self).__init__(parent=parent)
        self.setMouseTracking(True)
        self.setFocusPolicy(Qt.FocusPolicy.StrongFocus)

        self.setContentsMargins(10, 12, 0, 0)

        self._theme = application_service.get_app_theme()
        self._hover = Action.NONE
        self._press = Action.NONE

        self._pen = QPen(Qt.GlobalColor.transparent)

        self._brush = QBrush(Qt.GlobalColor.transparent)

        _top_offset = self.contentsMargins().top() if platform_service.is_macos else 0
        self._side_offset = (
            self.contentsMargins().left()
            if platform_service.is_macos
            else self.contentsMargins().right()
        )

        if platform_service.is_macos:
            _width = 13
            _height = 13
            _spacing = 7
        elif platform_service.is_windows:
            _width = 45
            _height = 35
            _spacing = 0
        else:
            _width = 17
            _height = 17
            _spacing = 8

        self._a_rect = QRect(self._side_offset, _top_offset, _width, _height)
        self._b_rect = self._a_rect.translated(_width + _spacing, 0)
        self._c_rect = self._b_rect.translated(_width + _spacing, 0)

        if platform_service.is_macos:
            self.container_height = (
                self._a_rect.y() + self._a_rect.height() + self.contentsMargins().top()
            )
        elif platform_service.is_windows:
            self.container_height = self._a_rect.y() + self._a_rect.height()
        else:
            self.container_height = (
                self._a_rect.y()
                + self._a_rect.height()
                + self.contentsMargins().top() * 2
            )

        self.container_width = self._c_rect.x() + self._c_rect.width()

        self.setFixedHeight(self.container_height)
        self.setFixedWidth(self.container_width)

        _rect_list = [
            self._a_rect,
            self._b_rect,
            self._c_rect,
        ]

        _button_data = [
            ButtonData(
                self._theme.win_minimize,
                Action.MINIMIZE,
                self.minimize_clicked,
                self._draw_min,
                lambda: (
                    self.parent().windowFlags() & Qt.WindowType.WindowMinimizeButtonHint
                )
                > 0,
            ),
            ButtonData(
                self._theme.win_maximize,
                Action.MAXIMIZE,
                self.maximize_clicked,
                self._draw_max,
                lambda: (
                    self.parent().windowFlags() & Qt.WindowType.WindowMaximizeButtonHint
                )
                > 0,
            ),
            ButtonData(
                self._theme.win_close,
                Action.CLOSE,
                self.close_clicked,
                self._draw_close,
                lambda: (
                    self.parent().windowFlags() & Qt.WindowType.WindowCloseButtonHint
                )
                > 0,
            ),
        ]

        _order = [2, 0, 1] if platform_service.is_macos else [0, 1, 2]

        self._rect_map = []
        for i, o in enumerate(_order):
            self._rect_map.append((_rect_list[i], _button_data[o]))

    def paintEvent(self, event: QPaintEvent) -> None:
        event.accept()

        p = QPainter(self)
        p.setRenderHint(QPainter.RenderHint.Antialiasing)

        for rect, data in self._rect_map:
            enabled = self.isActiveWindow() and data.enabled_fn()

            self._draw_button(p, rect, data.color, data.action, enabled)

            if platform_service.is_macos and not enabled:
                pass
            else:
                c_shape = data.color if enabled else data.color.darker(300)
                c_shape = c_shape.darker(200) if platform_service.is_macos else c_shape
                data.shape_fn(p, rect, c_shape, data.action)

        p.end()

    def mouseMoveEvent(self, event: QMouseEvent) -> None:
        prev = self._hover
        for rect, data in self._rect_map:
            if not data.enabled_fn():
                continue

            if rect.contains(event.pos()):
                self._hover = data.action
                break
        else:
            self._hover = Action.NONE

        if self._hover != prev:
            # repaint only on hover change
            self.update()

        super().mouseMoveEvent(event)

    def leaveEvent(self, event: QEvent) -> None:
        self._hover = Action.NONE
        self.update()

        super().leaveEvent(event)

    def mouseReleaseEvent(self, event: QMouseEvent) -> None:
        for rect, data in self._rect_map:
            if not data.enabled_fn():
                continue

            if rect.contains(event.pos()):
                data.signal.emit()
                break

        self._press = Action.NONE

        self.update()

        super().mouseReleaseEvent(event)

    def mousePressEvent(self, event: QMouseEvent) -> None:
        for rect, data in self._rect_map:
            if not data.enabled_fn():
                continue

            if rect.contains(event.pos()):
                self._press = data.action
                break
        else:
            self._press = Action.NONE

        self.update()

        super().mousePressEvent(event)

    def resizeEvent(self, event: QResizeEvent) -> None:
        if platform_service.is_macos:
            pass
        elif platform_service.is_windows:
            win_width = self.window().width()
            self.setGeometry(win_width - self.width(), 0, self.width(), self.height())
        else:
            win_width = self.window().width()
            self.setGeometry(
                win_width - self.width() - self.contentsMargins().left(),
                self.contentsMargins().top(),
                self.width(),
                self.height(),
            )

        super().resizeEvent(event)

    def _get_button_color(
        self, color: QColor, action: Action, is_enabled: bool
    ) -> QColor:
        if platform_service.is_windows:
            return QColor(Qt.GlobalColor.transparent)
        else:
            return color if is_enabled else self._theme.win_action_inactive

    def _draw_button(
        self, p: QPainter, rect: QRect, color: QColor, action: Action, enabled: bool
    ):
        p.setPen(Qt.GlobalColor.transparent)

        f_col = self._get_button_color(color, action, enabled)

        if enabled:
            if self._hover == action:
                f_col = color
            if self._press == action:
                f_col = color.lighter(135)

        p.setBrush(f_col)

        if platform_service.is_windows:
            if action == Action.CLOSE:
                offset = QPoint(0, 1)
                path = QPainterPath(rect.topLeft() + offset)
                path.lineTo(rect.bottomLeft() + offset)
                path.lineTo(rect.bottomRight() + offset)

                a = rect.topRight() + offset + QPoint(0, self._theme.win_corner_radius)
                b = rect.topRight() + offset - QPoint(self._theme.win_corner_radius, 0)

                path.lineTo(a)
                path.cubicTo(rect.topRight(), b, b)

                path.lineTo(rect.topLeft() + offset)
                p.drawPath(path)
            else:
                p.drawRect(rect)
        else:
            p.drawEllipse(rect)

    def _get_shape_tools(self, color: QColor, action: Action) -> Tuple[QPen, QBrush]:
        if self._press == action:
            color = color if platform_service.is_macos else color.lighter(200)
        elif self._hover == action:
            if platform_service.is_macos:
                color = color
            elif platform_service.is_windows:
                color = color.lighter(200)
            else:
                color = color.lighter()
        else:
            if platform_service.is_macos:
                color = Qt.GlobalColor.transparent
            elif platform_service.is_windows:
                color = Qt.GlobalColor.gray

        pen = QPen(color)
        pen.setWidth(3 if platform_service.is_windows else 2)
        pen.setCapStyle(Qt.PenCapStyle.RoundCap)

        brush = QBrush(color)

        return pen, brush

    def _scale_rect(self, rect: QRect, sw: int | float, sh: int | float) -> QRect:
        w = rect.width() * sh
        h = rect.height() * sw

        return rect.adjusted(w, h, -w, -h)

    def _draw_close(self, p: QPainter, rect: QRect, color: QColor, action: Action):
        pen, brush = self._get_shape_tools(color, action)
        p.setPen(pen)

        oH = (
            0.3
            if platform_service.is_macos
            else 0.325
            if platform_service.is_windows
            else 0.35
        )
        oW = (
            0.3
            if platform_service.is_macos
            else 0.375
            if platform_service.is_windows
            else 0.35
        )

        rect = self._scale_rect(rect.adjusted(1, 1, 0, 0), oH, oW)

        p.drawLines(
            [
                QLine(rect.topLeft(), rect.bottomRight()),
                QLine(rect.bottomLeft(), rect.topRight()),
            ]
        )

    def _draw_max(self, p: QPainter, rect: QRect, color: QColor, action: Action):
        pen, brush = self._get_shape_tools(color, action)
        p.setPen(pen)

        oH = (
            0.3
            if platform_service.is_macos
            else 0.325
            if platform_service.is_windows
            else 0.25
        )
        oW = (
            0.3
            if platform_service.is_macos
            else 0.375
            if platform_service.is_windows
            else 0.25
        )

        rect = self._scale_rect(rect, oH, oW)
        if platform_service.is_macos:
            p.setPen(Qt.GlobalColor.transparent)
            p.setBrush(brush)
            top = QRect(rect)
            p.drawPolygon([top.topLeft(), top.bottomLeft(), top.topRight()])
            bottom = rect.adjusted(1, 1, 1, 1)
            p.drawPolygon(
                [bottom.topRight(), bottom.bottomLeft(), bottom.bottomRight()]
            )
        else:
            p.drawRoundedRect(rect, 1, 1)

    def _draw_min(self, p: QPainter, rect: QRect, color: QColor, action: Action):
        pen, brush = self._get_shape_tools(color, action)
        p.setPen(pen)

        oH = (
            0.3
            if platform_service.is_macos
            else 0.325
            if platform_service.is_windows
            else 0.25
        )
        oW = (
            0.3
            if platform_service.is_macos
            else 0.375
            if platform_service.is_windows
            else 0.25
        )

        rect = self._scale_rect(rect, oH, oW)
        div = 2 if platform_service.is_macos or platform_service.is_windows else 1
        half = rect.y() + (rect.height() / div)
        p.drawLine(QLine(rect.x(), half, rect.x() + rect.width(), half))



from typing import TypeAlias

from PySide6.QtCore import Qt, Signal
from PySide6.QtGui import QIcon
from PySide6.QtWidgets import (
    QAbstractButton,
    QDialog,
    QDialogButtonBox,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QVBoxLayout,
    QWidget,
)

from pyside_app_core.frameless.base_window import FramelessBaseMixin

StandardButton: TypeAlias = QDialogButtonBox.StandardButton
ButtonRole: TypeAlias = QDialogButtonBox.ButtonRole


class FramelessBaseDialog(FramelessBaseMixin, QDialog):
    clicked = Signal(QAbstractButton)

    def __init__(self, icon: QIcon | None = None):
        super().__init__(parent=None, f=Qt.WindowType.WindowCloseButtonHint)

        self.layout().setContentsMargins(10, 0, 10, 10)
        self.layout().addSpacing(20)

        _margin = 20 if icon else 0

        _div = QHBoxLayout()
        _div.setContentsMargins(_margin, 0, _margin, 0)
        self.layout().addLayout(_div)

        if icon:
            _icon = QLabel(parent=self)
            _icon.setFixedSize(64, 64)
            _icon.setPixmap(icon.pixmap(64, 64))
            _div.addWidget(_icon, alignment=Qt.AlignmentFlag.AlignTop)

        self._info = QVBoxLayout()
        self._info.setContentsMargins(_margin, 0, 0, 0)
        _div.addLayout(self._info)

        self._button_box = QDialogButtonBox()
        self._button_box.rejected.connect(self.reject)
        self._button_box.accepted.connect(self.accept)
        self._button_box.clicked.connect(self.clicked.emit)

        self.layout().addWidget(self._button_box)

    def addWidget(self, widget: QWidget, *args, **kwargs) -> None:
        self._info.addWidget(widget, *args, **kwargs)

    def addStretch(self, val: int = 0):
        self._info.addStretch(val)

    def setStandardButtons(self, buttons: StandardButton):
        self._button_box.setStandardButtons(buttons)

    def setDefaultButton(self, button: StandardButton):
        std_button = self.button(button)
        std_button.setDefault(True)

    def setButtonText(self, button: StandardButton, text: str):
        std_button = self.button(button)
        std_button.setText(text)

    def button(self, button: StandardButton) -> QPushButton:
        return self._button_box.button(button)

    def buttonRole(self, button: QAbstractButton) -> ButtonRole:
        return self._button_box.buttonRole(button)



import inspect
import logging
from typing import Dict

from pyside_app_core.log.logger import PACLogger

__logger_cache: Dict[str, logging.Logger] = {}

logging.setLoggerClass(PACLogger)

__pac_name = "pyside_app_core"


def _get_caller_name():
    try:
        stack = inspect.stack()
        frame = stack[2][0]
        module = inspect.getmodule(frame)
        if not module:
            return __pac_name
    except:
        return __pac_name
    name = module.__name__
    if name == "__main__":
        name = __pac_name

    return name


def _get_cached_logger(name):
    if name in __logger_cache:
        lg = __logger_cache[name]
    else:
        lg = logging.getLogger(name)
        __logger_cache[name] = lg

    return lg


def set_level(name: str | None = None, lvl: int | None = None):
    """set the logger level"""
    if not name:
        name = _get_caller_name()
    if not lvl:
        raise ValueError(f"Please set '{name}' logger level explicitly.")
    lg = _get_cached_logger(name)
    lg.setLevel(lvl)


def debug(msg, *args, **kwargs):
    """call logger's debug"""
    name = _get_caller_name()
    lg = _get_cached_logger(name)
    lg.debug(msg, *args, **kwargs)


def info(msg, *args, **kwargs):
    """call logger's info"""
    name = _get_caller_name()
    lg = _get_cached_logger(name)
    lg.info(msg, *args, **kwargs)


def warning(msg, *args, **kwargs):
    """call logger's warning"""
    name = _get_caller_name()
    lg = _get_cached_logger(name)
    lg.warning(msg, *args, **kwargs)


def error(msg, *args, **kwargs):
    """call logger's error"""
    name = _get_caller_name()
    lg = _get_cached_logger(name)
    lg.error(msg, *args, **kwargs)


def critical(msg, *args, **kwargs):
    """call logger's critical"""
    name = _get_caller_name()
    lg = _get_cached_logger(name)
    lg.critical(msg, *args, **kwargs)


def exception(msg, *args, **kwargs):
    """call logger's exception"""
    name = _get_caller_name()
    lg = _get_cached_logger(name)
    lg.exception(msg, *args, **kwargs)


def set_all_level(lvl: int):
    """set level of all loggers"""
    for key in sorted(__logger_cache.keys()):
        lg = __logger_cache[key]
        if not isinstance(lg, PACLogger):
            continue
        lg.setLevel(lvl)



from pathlib import Path
from typing import List

from pyside_app_core.theme_generator.style_types import QtResourceFile, QtResourceGroup

__resources = Path(__file__).parent.parent / "resources"


STANDARD_RESOURCES: List[QtResourceGroup] = [
    QtResourceGroup(files=[QtResourceFile("style.qss")]),  # type: ignore[call-arg]
    QtResourceGroup(  # type: ignore[call-arg]
        prefix="std/",
        files=[
            QtResourceFile(
                str(__resources / "icons" / "console.svg"), alias="icons/console"
            ),
            QtResourceFile(
                str(__resources / "icons" / "down-arrow.svg"), alias="icons/down-arrow"
            ),
            QtResourceFile(
                str(__resources / "icons" / "grab-corner.svg"),
                alias="icons/grab-corner",
            ),
            QtResourceFile(
                str(__resources / "icons" / "reload.svg"), alias="icons/reload"
            ),
            QtResourceFile(str(__resources / "icons" / "save.svg"), alias="icons/save"),
            QtResourceFile(
                str(__resources / "icons" / "spin-down/normal.svg"),
                alias="icons/spin-down/normal",
            ),
            QtResourceFile(
                str(__resources / "icons" / "spin-down/disabled.svg"),
                alias="icons/spin-down/disabled",
            ),
            QtResourceFile(
                str(__resources / "icons" / "spin-up/normal.svg"),
                alias="icons/spin-up/normal",
            ),
            QtResourceFile(
                str(__resources / "icons" / "spin-up/disabled.svg"),
                alias="icons/spin-up/disabled",
            ),
        ],
    ),
]



from typing import cast

from PySide6.QtCore import QPoint, QRect, QSize, Qt, Signal
from PySide6.QtGui import (
    QBrush,
    QMouseEvent,
    QPainter,
    QPainterPath,
    QPaintEvent,
    QPen,
    QResizeEvent,
)
from PySide6.QtWidgets import QLabel, QMainWindow, QSizePolicy, QVBoxLayout, QWidget

from pyside_app_core.frameless.window_actions import WindowActions
from pyside_app_core.frameless.window_shade import FramelessWindowShade
from pyside_app_core.qt import application_service


class FramelessBaseMixin:
    close_window = Signal()

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        super(FramelessBaseMixin, self).__init__(*args, **kwargs)

        cast(QWidget, self).setWindowFlags(
            cast(QWidget, self).windowFlags() | Qt.WindowType.FramelessWindowHint
        )
        cast(QWidget, self).setAttribute(Qt.WidgetAttribute.WA_TranslucentBackground)
        cast(QWidget, self).setMouseTracking(True)
        cast(QWidget, self).setMinimumSize(300, 128)

        self._theme = application_service.get_app_theme()

        self._top_layout = QVBoxLayout()
        self._top_layout.setContentsMargins(5, 0, 5, 5)
        self._title = QLabel("", parent=cast(QWidget, self))
        self._title.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self._title.setSizePolicy(
            QSizePolicy.Policy.Fixed,
            QSizePolicy.Policy.MinimumExpanding,
        )
        self._title.setFixedHeight(32)
        self._top_layout.addWidget(
            self._title,
            alignment=Qt.AlignmentFlag.AlignTop,
        )

        cast(QWidget, self).setLayout(self._top_layout)

        self._moving = False
        self._screen_number: int = 0
        self._move_offset = QPoint(0, 0)
        self._move_corner = cast(QWidget, self).window().rect().bottomRight()
        self._can_minimize = True
        self._can_maximize = True

        self._window_actions = WindowActions(parent=cast(QWidget, self))
        self._window_actions.close_clicked.connect(cast(QWidget, self).close)
        self._window_actions.minimize_clicked.connect(self._on_minimize)
        self._window_actions.maximize_clicked.connect(self._on_maximize)

        self._window_shade = FramelessWindowShade(parent=cast(QWidget, self))

    @property
    def window_bar_geo(self) -> QRect:
        return QRect(
            0,
            0,
            cast(QWidget, self).rect().width(),
            self._window_actions.geometry().height(),
        )

    def setWindowTitle(self, title: str):
        self._title.setText(title)
        cast(QWidget, super()).setWindowTitle(title)

    def layout(self) -> QVBoxLayout:
        return cast(QWidget, super()).layout()

    def setFixedSize(self, *args):
        if not args:
            raise ValueError

        if len(args) == 1 and isinstance(args[0], QSize):
            cast(QWidget, super()).setFixedSize(args[0])
            self._can_maximize = False
            return

        elif len(args) == 2 and isinstance(args[0], int) and isinstance(args[1], int):
            cast(QWidget, super()).setFixedSize(args[0], args[1])
            self._can_maximize = False
            return

        raise ValueError

    def paintEvent(self, event: QPaintEvent) -> None:
        event.accept()

        p = QPainter(cast(QWidget, self))
        p.setRenderHint(QPainter.RenderHint.Antialiasing)

        p.setPen(Qt.GlobalColor.transparent)
        p.setBrush(self._theme.background_color)

        p.drawRoundedRect(
            cast(QWidget, self).rect(),
            self._theme.win_corner_radius,
            self._theme.win_corner_radius,
        )

        self._draw_border(p, cast(QWidget, self).rect())

        # TODO: a bit crude - we want the border to be on top of everything, but no need to call
        #  every time - how to call after all subclasses have added edited widgets? unknown...
        self._window_shade.raise_()
        self._window_actions.raise_()

        p.end()

    def resizeEvent(self, event: QResizeEvent) -> None:
        self._title.setFixedWidth(cast(QWidget, self).rect().width())
        self._window_actions.resizeEvent(event)
        cast(QWidget, super()).resizeEvent(event)

    def _draw_border(self, p: QPainter, rect: QRect):
        # outline
        p.setBrush(QBrush(Qt.GlobalColor.transparent))

        draw_rect = rect.adjusted(1, 1, -1, -1)

        top_corner = QPainterPath(rect.topRight())
        top_corner.lineTo(rect.topLeft())
        top_corner.lineTo(rect.bottomLeft())
        p.setClipPath(top_corner)

        p.setPen(QPen(self._theme.background_color.lighter(220), 1))
        p.drawRoundedRect(
            draw_rect,
            self._theme.win_corner_radius - 2,
            self._theme.win_corner_radius - 2,
        )

        bottom_corner = QPainterPath(rect.topRight())
        bottom_corner.lineTo(rect.bottomRight())
        bottom_corner.lineTo(rect.bottomLeft())
        p.setClipPath(bottom_corner)

        p.setPen(QPen(self._theme.background_color.lighter(130), 2))
        p.drawRoundedRect(
            draw_rect,
            self._theme.win_corner_radius - 2,
            self._theme.win_corner_radius - 2,
        )
        p.setClipping(False)

    def _on_maximize(self):
        if cast(QWidget, self).isMaximized():
            cast(QWidget, self).setWindowState(Qt.WindowState.WindowNoState)
        else:
            cast(QWidget, self).setWindowState(Qt.WindowState.WindowMaximized)

    def _on_minimize(self):
        cast(QWidget, self).setWindowState(Qt.WindowState.WindowMinimized)

    def mouseMoveEvent(self, event: QMouseEvent) -> None:
        if self._moving:
            win = cast(QWidget, self).window()
            pos = win.mapToGlobal(event.pos())
            target = pos - self._move_offset
            win.move(target)

            cast(QWidget, self).update()

        cast(QWidget, super()).mouseMoveEvent(event)

    def mousePressEvent(self, event: QMouseEvent) -> None:
        if (
            (event.buttons() & Qt.MouseButton.LeftButton) == Qt.MouseButton.LeftButton
            and self.window_bar_geo.contains(event.pos())
            and not self._window_actions.geometry().contains(event.pos())
        ):
            self._moving = True

            win = cast(QWidget, self).window()
            rect = win.rect()
            pos: QPoint = win.mapToGlobal(event.pos())
            corner: QPoint = win.mapToGlobal(rect.topLeft())
            self._move_offset = pos - corner
            self._move_corner = rect.bottomRight()

        cast(QWidget, super()).mousePressEvent(event)

    def mouseReleaseEvent(self, event: QMouseEvent) -> None:
        self._moving = False
        cast(QWidget, self).update()

        cast(QWidget, super()).mouseReleaseEvent(event)

    def mouseDoubleClickEvent(self, event: QMouseEvent) -> None:
        self._on_maximize()

        cast(QWidget, super()).mouseReleaseEvent(event)



from PySide6.QtCore import Qt
from PySide6.QtGui import QBrush, QPainter, QPaintEvent, QPen
from PySide6.QtWidgets import QWidget

from pyside_app_core.qt import application_service


class FramelessWindowShade(QWidget):
    def __init__(
        self,
        parent: QWidget,
    ):
        super(FramelessWindowShade, self).__init__(parent=parent)

        self.setAttribute(Qt.WidgetAttribute.WA_TransparentForMouseEvents)

        self._theme = application_service.get_app_theme()
        self._disabled = False

    def setDisabled(self, val: bool) -> None:
        self._disabled = val

    def setEnabled(self, val: bool) -> None:
        self.setDisabled(not val)

    def paintEvent(self, event: QPaintEvent) -> None:
        event.accept()

        rect = self.parent().contentsRect()
        self.setFixedSize(rect.size())

        p = QPainter(self)
        p.setRenderHint(QPainter.RenderHint.Antialiasing)
        p.setPen(QPen(Qt.GlobalColor.transparent))
        p.setBrush(QBrush(Qt.GlobalColor.transparent))

        if not self.isActiveWindow():
            p.setPen(QPen(Qt.GlobalColor.transparent))
            bkgd_color = self._theme.background_color.darker(125)
            bkgd_color.setAlpha(80)
            p.setBrush(QBrush(bkgd_color))
            p.drawRoundedRect(
                rect, self._theme.win_corner_radius, self._theme.win_corner_radius
            )

        p.end()



"""
Adapted from:
https://gis.stackexchange.com/questions/350148/qcombobox-multiple-selection-pyqt5
"""

from typing import Generator, Generic, Tuple, TypeVar

from PySide6.QtCore import QEvent, Qt, Signal
from PySide6.QtGui import QFontMetrics, QStandardItem
from PySide6.QtWidgets import QComboBox, QStyledItemDelegate, QWidget

from pyside_app_core.qt.widgets.object_name_mixin import ObjectNameMixin
from pyside_app_core.qt.widgets.settings_mixin import SettingsMixin

DT = TypeVar("DT")


class MultiComboBox(ObjectNameMixin, SettingsMixin, Generic[DT], QComboBox):
    # Subclass Delegate to increase item height
    class Delegate(QStyledItemDelegate):
        def sizeHint(self, option, index):
            size = super().sizeHint(option, index)
            size.setHeight(20)
            return size

    selectionChanged = Signal(list)

    def __init__(self, parent: QWidget, *args, **kwargs):
        super(MultiComboBox, self).__init__(parent=parent, *args, **kwargs)

        # Make the combo editable to set a custom text, but readonly
        self.setEditable(True)
        self.lineEdit().setReadOnly(True)
        self.lineEdit().setPlaceholderText("Command Filter")

        # Use custom delegate
        self.setItemDelegate(MultiComboBox.Delegate())

        # Hide and show popup when clicking the line edit
        self.lineEdit().installEventFilter(self)
        self.closeOnLineEditClick = False

        # Prevent popup from closing when clicking on an item
        self.view().viewport().installEventFilter(self)

        self.addItem("Show/Hide All", isChecked=True)

        self.model().itemChanged.connect(self._update_selection)
        self.model().dataChanged.connect(self._update_text)
        self.model().dataChanged.connect(self._emit_current_data)

    def _emit_current_data(self):
        self.selectionChanged.emit(self.currentData())

    def _update_selection(self, item: QStandardItem):
        if item.index().row() == 0:
            state = item.checkState()
            self.model().layoutAboutToBeChanged.emit()
            for item in self.modelItems():
                item.setCheckState(state)
            self.model().layoutChanged.emit()

    def resizeEvent(self, event):
        # Recompute text to elide as needed
        self._update_text()
        super().resizeEvent(event)

    def eventFilter(self, object, event):
        if object == self.lineEdit():
            if event.type() == QEvent.Type.MouseButtonRelease:
                if self.closeOnLineEditClick:
                    self.hidePopup()
                else:
                    self.showPopup()
                return True
            return False

        if object == self.view().viewport():
            if event.type() == QEvent.Type.MouseButtonRelease:
                index = self.view().indexAt(event.pos())
                item = self.model().item(index.row())

                if item.checkState() == Qt.CheckState.Checked:
                    item.setCheckState(Qt.CheckState.Unchecked)
                else:
                    item.setCheckState(Qt.CheckState.Checked)
                return True
        return False

    def showPopup(self):
        super().showPopup()
        # When the popup is displayed, a click on the lineedit should close it
        self.closeOnLineEditClick = True

    def hidePopup(self):
        super().hidePopup()
        # Used to prevent immediate reopening when clicking on the lineEdit
        self.startTimer(100)
        # Refresh the display text when closing
        self._update_text()

    def timerEvent(self, event):
        # After timeout, kill timer, and re-enable click on line edit
        self.killTimer(event.timerId())
        self.closeOnLineEditClick = False

    def _update_text(self):
        texts = []
        for item in self.modelItems():
            if item.checkState() == Qt.CheckState.Checked:
                texts.append(item.text())
        text = ", ".join(texts)

        # Compute elided text (with "...")
        metrics = QFontMetrics(self.lineEdit().font())
        text = metrics.elidedText(
            text, Qt.TextElideMode.ElideRight, self.lineEdit().width()
        )
        self.lineEdit().setText(text)

    def addItem(self, text, userData: DT | None = None, isChecked=False, **kwargs):
        item = QStandardItem()
        item.setText(text)
        if userData is None:
            item.setData(text, Qt.ItemDataRole.UserRole)
        else:
            item.setData(userData, Qt.ItemDataRole.UserRole)

        item.setFlags(Qt.ItemFlag.ItemIsEnabled | Qt.ItemFlag.ItemIsUserCheckable)
        item.setCheckState(
            Qt.CheckState.Checked if isChecked else Qt.CheckState.Unchecked
        )

        self.model().appendRow(item)

    def addItems(
        self, texts: list[str], dataList: list[DT] | None = None, allChecked=False
    ):
        for i, text in enumerate(texts):
            try:
                data = (dataList or [])[i]
            except (TypeError, IndexError):
                data = None
            self.addItem(text, data, isChecked=allChecked)

    def currentData(self, role: Qt.ItemDataRole = Qt.ItemDataRole.UserRole) -> list[DT]:
        res = []
        for item in self.modelItems():
            if item.checkState() == Qt.CheckState.Checked:
                res.append(item.data(role))
        return res

    def currentOptions(self) -> list[Tuple[str, DT]]:
        res = []
        for item in self.modelItems():
            if item.checkState() == Qt.CheckState.Checked:
                res.append((item.text(), item.data(Qt.ItemDataRole.UserRole)))
        return res

    def modelItems(self) -> Generator[QStandardItem, None, None]:
        for i in range(1, self.model().rowCount()):
            yield self.model().item(i)

    def _restore_state(self) -> None:
        options = self._settings.value(f"{self.obj_name}_options")
        if not options:
            return

        for item in self.modelItems():
            state = (
                Qt.CheckState.Checked
                if item.text() in options
                else Qt.CheckState.Unchecked
            )
            item.setCheckState(state)

    def _store_state(self) -> None:
        self._settings.setValue(
            f"{self.obj_name}_options", [n for n, _ in self.currentOptions()]
        )



from dataclasses import dataclass, field
from pathlib import Path
from typing import List


@dataclass(frozen=True)
class QtResourceFile:
    path: str | Path
    alias: str | None = field(default=None)


@dataclass(frozen=True, kw_only=True)
class QtResourceGroup:
    prefix: str | None = field(default=None)
    lang: str | None = field(default=None)
    files: List[QtResourceFile] = field(default_factory=list)



from typing import Self

from PySide6.QtGui import QColor


class SColor(QColor):
    def __init__(self, *args, **kwargs):
        if (
            len(args) == 1
            and not kwargs
            and (isinstance(args[0], int) or isinstance(args[0], float))
        ):
            val = args[0]
            super(SColor, self).__init__(val, val, val)
        else:
            super(SColor, self).__init__(*args, **kwargs)

    def lighter(self, f=150) -> Self:
        return self.__class__(super().lighter(f))

    def darker(self, f=150) -> Self:
        return self.__class__(super().darker(f))

    def desaturated(self, f=150) -> Self:
        hsv = self.toHsv()
        hsv.setHsv(
            hsv.hue(),
            min(max(hsv.saturation() * (2 - max(f, 0) / 100.0), 0), 255),
            hsv.value(),
        )
        return self.__class__(hsv.toRgb())

    def __str__(self):
        return f"rgb({self.red()},{self.green()},{self.blue()})"



from pyside_app_core.style.pixel_val import PixelVal
from pyside_app_core.style.s_color import SColor
from pyside_app_core.services import platform_service


class QssTheme:
    """
    Theme configuration object with default values.
    Will be used for both static qss style and dynamic widget style.

    Notes:
        - subclasses can override properties in the body of the class.
        - subclasses can define new properties for use in custom qss templates, etc.
        - The instance CAN NOT be edited during runtime since qss files are not regenerated.
    """

    background_color = SColor(60)
    background_color_inactive = SColor(70)

    text_color = SColor(250)
    text_color_inactive = SColor(200)
    text_color_disabled = SColor(120)

    win_corner_radius = PixelVal(14)
    widget_corner_radius = PixelVal(8)

    win_divider_width = PixelVal(2)
    win_divider_color = SColor(85)

    widget_accent_main_color = SColor(80, 160, 250)
    widget_accent_neutral_color = SColor(85)
    widget_border_color = SColor(95)
    widget_hover_color = SColor(95)

    input_background_color = SColor(95)
    input_background_color_inactive = SColor(85)

    tooltip_background_color = SColor(85)

    win_close = (
        SColor(232, 16, 32) if platform_service.is_windows else SColor(237, 107, 95)
    )
    win_minimize = SColor(120) if platform_service.is_windows else SColor(245, 191, 79)
    win_maximize = SColor(120) if platform_service.is_windows else SColor(98, 197, 84)
    win_action_inactive = SColor(85)

    def __setattr__(self, key, value):
        raise NotImplementedError(
            f'Can\'t update property "{self.__class__.__name__}.{key}". QssThemes are immutable.'
        )


DEFAULT_THEME = QssTheme()



from pathlib import Path

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QApplication, QMainWindow

from pyside_app_core.errors.basic_errors import ApplicationError
from pyside_app_core.qt.util.style_sheet import (
    apply_style_sheet,
    register_resource_file,
)


class BaseApp(QApplication):
    def __init__(self, resources_rcc: Path | None = None, *args, **kwargs):
        super(BaseApp, self).__init__(*args, **kwargs)
        self.setStyle("Fusion")
        self.setAttribute(
            Qt.ApplicationAttribute.AA_UseStyleSheetPropagationInWidgetStyles
        )

        register_resource_file(resources_rcc)

        apply_style_sheet(":/style.qss", self)

        # override in subclass
        self._main_window: QMainWindow = None

    def launch(self) -> int:
        if not self._main_window:
            raise ApplicationError(
                f"Must subclass {self.__class__.__name__} and define a main window"
            )

        self._main_window.show()
        self._main_window.devicePixelRatio()
        return self.exec()



from PySide6.QtGui import QIcon
from PySide6.QtWidgets import QLabel

from pyside_app_core.frameless.base_dialog import (
    FramelessBaseDialog,
    StandardButton,
)
from pyside_app_core.qt import application_service


class AboutDialog(FramelessBaseDialog):
    def __init__(self):
        super().__init__(icon=QIcon(":/std/icons/console"))

        self.setFixedSize(500, 300)
        self.setWindowTitle(f"About {application_service.get_app_name()}")

        self.setStandardButtons(StandardButton.Close)
        self.setDefaultButton(StandardButton.Close)

        _version = QLabel(f"Version {application_service.get_app_version()}")
        self.addWidget(_version)

        self.addStretch()



import sys
import traceback
from types import TracebackType
from typing import Protocol, Type

from PySide6.QtWidgets import QDialog

from pyside_app_core import log


class ErrorDialogInterface(Protocol):
    def __init__(self, message: str, details: str | None = None):
        ...

    def exec(self) -> None:
        ...


_error_dialog_class: Type[ErrorDialogInterface] | None = None


def install_excepthook(error_dialog: Type[QDialog] | None = None) -> None:
    global _error_dialog_class
    _error_dialog_class = error_dialog
    sys.excepthook = __custom_excepthook


def __custom_excepthook(
    etype: Type[BaseException], evalue: BaseException, tb: TracebackType | None
):
    formatted_exception_string = "".join(traceback.format_exception(etype, evalue, tb))

    if hasattr(evalue, "internal"):
        if evalue.internal:
            log.exception(formatted_exception_string)
            return

    if _error_dialog_class:
        _error_dialog_class(
            f"<{etype.__name__}>\n{evalue}", formatted_exception_string
        ).exec()



import subprocess
import tempfile
from pathlib import Path
from typing import List, Literal, Type

from jinja2 import ChoiceLoader, Environment, FileSystemLoader, PackageLoader

from pyside_app_core.theme_generator.standard_resources import STANDARD_RESOURCES
from pyside_app_core.theme_generator.style_types import QtResourceGroup
from pyside_app_core.style.theme import DEFAULT_THEME, QssTheme


def _compose_template_environment(*extra_template_files: Path) -> Environment:
    loaders = [PackageLoader("pyside_app_core.resources")]

    if extra_template_files:
        dirs = list(set([str(p.parent) for p in extra_template_files]))
        loaders.append(FileSystemLoader(dirs))

    return Environment(loader=ChoiceLoader(loaders), autoescape=False)


def _compile_qss_template(theme: QssTheme, qss_extra: List[Path] | None = None) -> str:
    qss_extra = qss_extra or []
    env = _compose_template_environment(*qss_extra)
    qss_template = env.get_template("style.qss.jinja2")
    return qss_template.render(theme=theme, qss_extra=[q.name for q in qss_extra])


def _compile_qrc_template(resources: List[QtResourceGroup]) -> str:
    env = _compose_template_environment()
    qrc_template = env.get_template("resources.qrc.jinja2")
    return qrc_template.render(qresources=resources)


def _write_file(target: Path, data: str) -> Path:
    with open(target, "w") as f:
        f.write(data)

    return target


ResourceFormat = Literal["python", "binary"]


def compile_qrc_to_resources(
    target_dir: Path,
    qss_theme: QssTheme | Type[QssTheme],
    rcc_format: ResourceFormat = "binary",
    qss_template_extra: List[Path] | None = None,
    resources_extra: List[QtResourceGroup] | None = None,
) -> Path:
    if not isinstance(qss_theme, QssTheme):
        qss_theme = qss_theme()

    with tempfile.TemporaryDirectory() as tempdir:
        tempdir = Path(tempdir)

        _write_file(
            tempdir / "style.qss",
            _compile_qss_template(qss_theme or DEFAULT_THEME, qss_template_extra),
        )

        resources = STANDARD_RESOURCES
        if resources_extra:
            resources.extend(resources_extra)

        qrc_file = _write_file(
            tempdir / "resources.qrc", _compile_qrc_template(resources)
        )

        file_name = "resources.py"

        rcc_args = []
        if rcc_format == "binary":
            file_name = "resources.rcc"
            rcc_args.append("--binary")

        file_target = target_dir / file_name

        subprocess.check_call(
            ["pyside6-rcc", *rcc_args, "-o", str(file_target), str(qrc_file)]
        )

    return file_target



from typing import Type

from pyside_app_core.errors.basic_errors import ApplicationError
from pyside_app_core.style.theme import QssTheme

__app_name: str
__app_id: str
__app_version: str
__app_theme: QssTheme


def set_app_name(app_name: str) -> None:
    global __app_name
    if __app_name:
        raise ApplicationError("Can't update app_name after its been set once.")

    __app_name = app_name


def set_app_id(app_id: str) -> None:
    global __app_id
    if __app_id:
        raise ApplicationError("Can't update app_id after its been set once.")

    __app_id = app_id


def set_app_version(app_version: str) -> None:
    global __app_version
    if __app_version:
        raise ApplicationError("Can't update app_version after its been set once.")

    __app_version = app_version


def set_app_theme(app_theme: QssTheme) -> None:
    global __app_theme
    if __app_theme:
        raise ApplicationError("Can't update app_theme after its been set once.")

    __app_theme = app_theme


def get_app_name() -> str:
    global __app_name
    if not __app_name:
        raise ApplicationError("app_name is unset")
    return __app_name


def get_app_id() -> str:
    global __app_id
    if not __app_id:
        raise ApplicationError("app_id is unset")
    return __app_id


def get_app_version() -> str:
    global __app_version
    if not __app_version:
        raise ApplicationError("app_version is unset")
    return __app_version


def get_app_theme() -> QssTheme:
    global __app_theme
    if not __app_theme:
        raise ApplicationError("app_theme is unset")
    return __app_theme



from PySide6.QtGui import QClipboard
from PySide6.QtWidgets import QApplication, QLabel, QPlainTextEdit

from pyside_app_core.frameless.base_dialog import (
    FramelessBaseDialog,
    StandardButton,
)


class ErrorDialog(FramelessBaseDialog):
    def __init__(self, message: str, detailed: str | None = None):
        super().__init__()

        self.setWindowTitle("Encountered An Unexpected Error")
        self.setMinimumWidth(640)

        self._clipboard = QClipboard(parent=self)

        self.setStandardButtons(
            StandardButton.Ignore | StandardButton.Reset | StandardButton.Abort
        )
        self.setDefaultButton(StandardButton.Ignore)

        self.setButtonText(StandardButton.Ignore, "Ignore")
        self.setButtonText(StandardButton.Reset, "Copy")
        self.setButtonText(StandardButton.Abort, "Quit")

        _copy_btn = self.button(StandardButton.Reset)
        _copy_btn.clicked.connect(self._copy_to_clipboard)

        _quit_btn = self.button(StandardButton.Abort)
        _quit_btn.clicked.connect(QApplication.instance().quit)

        _m_lab = QLabel("Message:")
        self.addWidget(_m_lab)
        self._message_box = QPlainTextEdit(message, parent=self)
        self._message_box.setReadOnly(True)
        self.addWidget(self._message_box)

        self._trace_box = None
        if detailed:
            _t_lab = QLabel("Details:")
            self.addWidget(_t_lab)
            self._trace_box = QPlainTextEdit(detailed, parent=self)
            self._trace_box.setReadOnly(True)
            self.addWidget(self._trace_box)

    def _copy_to_clipboard(self):
        if self._trace_box:
            self._clipboard.setText(
                f"{self._message_box.toPlainText()}\n\n{self._trace_box.toPlainText()}"
            )
        else:
            self._clipboard.setText(self._message_box.toPlainText())



import os
import platform

__valid_platforms = ["Darwin", "Windows", "Linux"]
__platform = None


def _platform():
    global __platform

    system_platform = platform.system()
    override = os.environ.get("PLATFORM_OVERRIDE", system_platform)

    if not __platform:
        __platform = override if override in __valid_platforms else system_platform

    return __platform


is_macos = _platform() == "Darwin"
is_windows = _platform() == "Windows"
is_linux = _platform() == "Linux"



import contextlib
from typing import ContextManager

from PySide6 import QtGui
from PySide6.QtGui import QAction
from PySide6.QtWidgets import (
    QHBoxLayout,
    QMenu,
    QMenuBar,
    QSizePolicy,
    QVBoxLayout,
    QWidget,
)

from pyside_app_core.style.pixel_val import PixelVal
from pyside_app_core.style.s_color import SColor
from pyside_app_core.qt.widgets.object_name_mixin import ObjectNameMixin


class MenuContext(QMenu):
    @contextlib.contextmanager
    def add_menu(self, name: str) -> ContextManager["MenuContext"]:
        menu = MenuContext(name, self)
        menu.setObjectName(f"Menu{name.replace(' ', '_')}")
        yield menu
        self.addMenu(menu)

    @contextlib.contextmanager
    def add_action(self, name: str) -> ContextManager[QAction]:
        action = QtGui.QAction(name, self)
        action.setObjectName(f"MenuAction{name.replace(' ', '_')}")
        yield action
        self.addAction(action)


class MenuBarContext(ObjectNameMixin, QWidget):
    def __init__(
        self, parent: QWidget, border_width=PixelVal(2), border_color=SColor(20, 20, 20)
    ):
        super(MenuBarContext, self).__init__(parent=parent)

        self._border_width = border_width
        self.setStyleSheet(
            f"""
            #{self.obj_name} {{
                border-bottom: {border_width} solid {border_color};
            }}
            """
        )

        self.setContentsMargins(0, 0, 0, 0)

        self._layout = QVBoxLayout()
        self._layout.setSpacing(0)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self._layout)

        self._layout.addStretch()

        self._menu_layout = QHBoxLayout()
        self._menu_layout.setSpacing(0)
        self._menu_layout.setContentsMargins(0, 0, 0, 0)
        self._layout.addLayout(self._menu_layout)

        self._shift = QWidget(self)
        self._menu_layout.addWidget(self._shift)
        self._shift.setFixedWidth(0)

        self._menu_bar = QMenuBar(parent=parent)
        self._menu_bar.setSizePolicy(
            QSizePolicy.Policy.Minimum, QSizePolicy.Policy.MinimumExpanding
        )
        self._menu_layout.addWidget(self._menu_bar)

        self._menu_layout.addStretch()

        self._layout.addStretch()

    def setNativeMenuBar(self, val: bool):
        self._menu_bar.setNativeMenuBar(val)

    @contextlib.contextmanager
    def add_menu(self, name: str) -> ContextManager[MenuContext]:
        menu = MenuContext(name, self)
        menu.setObjectName(f"Menu{name.replace(' ', '_')}")
        yield menu
        self._menu_bar.addMenu(menu)

    @contextlib.contextmanager
    def add_action(self, name: str) -> ContextManager[QAction]:
        action = QtGui.QAction(name, self)
        action.setObjectName(f"MenuBarAction{name.replace(' ', '_')}")
        yield action
        self._menu_bar.addAction(action)

    def set_offset(self, val: int):
        self.setFixedHeight(val + self._border_width)

    def set_shift(self, val: int):
        self._shift.setFixedWidth(val)
